import unittest
import numpy as np
from activations import sigmoid, sigmoid_derivative, relu, relu_derivative
from propagate import propagate, _cost, _step_forward, _propagate_forward, _step_backward, _propagate_back
from utils import vectorize, normalize_rows, softmax, loss
from initializations import initialize_parameters, initialize_lr_with_zeros


class TestActivations(unittest.TestCase):

  def test_sigmoid(self):
    self.assertEqual(sigmoid(0), 0.5)
    self.assertGreater(sigmoid(100), .99)
    self.assertLess(sigmoid(-100), .01)

    Z = np.array([1,2,3])
    expected = np.array([0.73105858, 0.88079708, 0.95257413])
    self.assertTrue(np.allclose(sigmoid(Z), expected))

  def test_sigmoid_derivative(self):
    self.assertEqual(sigmoid_derivative(0), 0.25)
    self.assertLess(sigmoid_derivative(100), .001)
    self.assertLess(sigmoid_derivative(-100), .001)

    Z = np.array([1, 2, 3])
    expected = np.array([0.1966119, 0.1049935, 0.04517666])
    self.assertTrue(np.allclose(sigmoid_derivative(Z), expected))

  def test_relu(self):
    self.assertEqual(relu(1), 1)
    self.assertEqual(relu(0), 0)
    self.assertEqual(relu(-1), 0)

  def test_relu_derivative(self):
    self.assertTrue(np.allclose(relu_derivative(np.array([[2,1,0,-1,-2]])), np.array([[1,1,0,0,0]])))

class TestUtils(unittest.TestCase):
  
  def test_vectorize(self):
    with self.assertRaises(AssertionError):
      vectorize([])
    
    array = vectorize(np.array([]))
    self.assertEqual(array.shape, (0,1))

    array = vectorize(np.zeros((2,1)))
    self.assertEqual(array.shape, (2,1))

    array = vectorize(np.zeros((1,3)))
    self.assertEqual(array.shape, (3,1))

    array = vectorize(np.zeros((3,4,2,5)))
    self.assertEqual(array.shape, (120,1))

    image = np.array([
      [
        [0.67826139, 0.29380381],
        [0.90714982, 0.52835647],
        [0.4215251 , 0.45017551]
      ],
      [
        [0.92814219, 0.96677647],
        [0.85304703, 0.52351845],
        [0.19981397, 0.27417313]
      ],
      [
        [0.60659855, 0.00533165],
        [0.10820313, 0.49978937],
        [0.34144279, 0.94630077]
      ]
    ])

    expected = np.array([
      [0.67826139],
      [0.29380381],
      [0.90714982],
      [0.52835647],
      [0.4215251 ],
      [0.45017551],
      [0.92814219],
      [0.96677647],
      [0.85304703],
      [0.52351845],
      [0.19981397],
      [0.27417313],
      [0.60659855],
      [0.00533165],
      [0.10820313],
      [0.49978937],
      [0.34144279],
      [0.94630077]
    ])

    self.assertTrue(np.allclose(vectorize(image), expected))

  def test_normalize_rows(self):
    arg = np.zeros((3,2))
    self.assertTrue(np.array_equal(normalize_rows(arg), arg))

    arg = np.array([
      [0, 3, 4],
      [1, 6, 4]
    ])

    expected = np.array([
      [0, 0.6, 0.8],
      [0.13736056, 0.82416338, 0.54944226]
    ])

    self.assertTrue(np.allclose(normalize_rows(arg), expected))

  def test_softmax(self):
    arg = np.zeros((3,2))
    self.assertTrue(np.array_equal(softmax(arg), arg + 0.5))

    arg = np.array([
      [9, 2, 5, 0, 0],
      [7, 5, 0, 0 ,0]
    ])

    expected = np.array([
      [9.80897665e-01, 8.94462891e-04, 1.79657674e-02, 1.21052389e-04, 1.21052389e-04],
      [8.78679856e-01, 1.18916387e-01, 8.01252314e-04, 8.01252314e-04, 8.01252314e-04]
    ])

    self.assertTrue(np.allclose(softmax(arg), expected))

  def test_loss(self):
    with self.assertRaises(AssertionError):
      loss([], [], L=0)
    
    with self.assertRaises(AssertionError):
      loss([], [], L=3)

    size = 10
    y = np.random.randint(2, size=size)
    yhat = np.copy(y)

    self.assertEqual(loss(yhat, y, L=1), 0)
    self.assertEqual(loss(yhat, y, L=2), 0)

    yhat = (y == 0).astype(int)
    self.assertEqual(loss(yhat, y, L=1), size)
    self.assertEqual(loss(yhat, y, L=2), size)

    y = np.array([1, 0, 0, 1, 1])
    yhat = np.array([.9, 0.2, 0.1, .4, .9])

    self.assertEqual(loss(yhat, y, L=1), 1.1)
    self.assertEqual(loss(yhat, y, L=2), 0.43)

class TestInitialization(unittest.TestCase):

  def test_initialize_parameters(self):
    for i in range(5, 8):
      for o in range(1, 4):
        layer_dimensions = [dimension for dimension in [i,o] if dimension != 0]
        parameters = initialize_parameters(layer_dimensions, use="zeros")
        self.assertEqual(len(parameters), 2)
        
        self.assertTrue(np.array_equal(parameters["W1"], np.zeros((o,i))))
        self.assertTrue(np.array_equal(parameters["b1"], np.zeros((o, 1))))
        for h1 in range(1, 4):
          layer_dimensions = [dimension for dimension in [i, h1, o] if dimension != 0]
          parameters = initialize_parameters(layer_dimensions, use="zeros")
          self.assertEqual(len(parameters), 4)
          
          self.assertTrue(np.array_equal(parameters["W1"], np.zeros((h1,i))))
          self.assertTrue(np.array_equal(parameters["b1"], np.zeros((h1, 1))))
          
          self.assertTrue(np.array_equal(parameters["W2"], np.zeros((o, h1))))
          self.assertTrue(np.array_equal(parameters["b2"], np.zeros((o, 1))))
          for h2 in range(7, 9):
            layer_dimensions = [dimension for dimension in [i,h1,h2,o] if dimension != 0]
            parameters = initialize_parameters(layer_dimensions, use="zeros")
            self.assertEqual(len(parameters), 6)
            
            self.assertTrue(np.array_equal(parameters["W1"], np.zeros((h1,i))))
            self.assertTrue(np.array_equal(parameters["b1"], np.zeros((h1, 1))))
            
            self.assertTrue(np.array_equal(parameters["W2"], np.zeros((h2, h1))))
            self.assertTrue(np.array_equal(parameters["b2"], np.zeros((h2, 1))))
            
            self.assertTrue(np.array_equal(parameters["W3"], np.zeros((o, h2))))
            self.assertTrue(np.array_equal(parameters["b3"], np.zeros((o, 1))))

    np.random.seed(3)
    expected = {
      'W1': np.array([
        [17.88628473, 4.36509851, 0.96497468],
        [-18.63492703, -2.77388203, -3.54758979]
      ]),
      'b1': np.array([[0], [0]]),
      'W2': np.array([[-0.82741481, -6.27000677]]),
      'b2': np.array([[0]])
    }

    parameters = initialize_parameters([3,2,1], use="random", W_multiplier=10)
    
    self.assertTrue(np.allclose(parameters["W1"], expected['W1']))
    self.assertTrue(np.allclose(parameters["b1"], expected['b1']))
    self.assertTrue(np.allclose(parameters["W2"], expected['W2']))
    self.assertTrue(np.allclose(parameters["b2"], expected['b2']))

    np.random.seed(3)
    expected = {
      'W1': np.array([
        [1.78862847, 0.43650985],
        [0.09649747, -1.8634927],
        [-0.2773882, -0.35475898],
        [-0.08274148, -0.62700068]
      ]),
      'b1': np.array([
        [0.],
        [0.],
        [0.],
        [0.]
      ]),
      'W2': np.array([[-0.03098412, -0.33744411, -0.92904268, 0.62552248]]),
      'b2': np.array([[0]])
    }

    parameters = initialize_parameters([2,4,1], use="he")
    
    self.assertTrue(np.allclose(parameters["W1"], expected['W1']))
    self.assertTrue(np.allclose(parameters["b1"], expected['b1']))
    print(parameters['W2'] / 2)
    self.assertTrue(np.allclose(parameters["W2"], expected['W2']))
    self.assertTrue(np.allclose(parameters["b2"], expected['b2']))
    

  def test_initialize_lr_with_zeros(self):
    with self.assertRaises(ValueError):
      initialize_lr_with_zeros(0)

    for l in range(1,4):
      parameters = initialize_lr_with_zeros(l)
      self.assertTrue(np.array_equal(parameters['W1'], np.zeros((1,l))))
      self.assertEqual(parameters['b1'], 0)
      self.assertEqual(parameters['b1'].shape, (1, 1))

class TestPropagate(unittest.TestCase):

  def test__step_forward(self):
    A_prev = np.array([
      [-0.41675785, -0.05626683],
      [-2.1361961, 1.64027081],
      [-1.79343559, -0.84174737]
    ])
    W = np.array([[0.50288142, -1.24528809, -1.05795222]])
    b = np.array([[-0.90900761]])

    activation = 'sigmoid'
    A, ((cA_prev, cW, cb), (cZ, cActivation)) = _step_forward(A_prev, W, b, activation)
    self.assertTrue(np.allclose(A, [[0.96890023, 0.11013289]]))
    self.assertIs(A_prev, cA_prev)
    self.assertIs(W, cW)
    self.assertIs(b, cb)
    self.assertTrue(np.allclose(cZ, np.array([[3.43896134, -2.08938436]])))
    self.assertIs(activation, cActivation)

    activation = 'relu'
    A, ((cA_prev, cW, cb), (cZ, cActivation)) = _step_forward(A_prev, W, b, activation)
    self.assertTrue(np.allclose(A, [[3.43896131, 0.]]))
    self.assertIs(A_prev, cA_prev)
    self.assertIs(W, cW)
    self.assertIs(b, cb)
    self.assertTrue(np.allclose(cZ, np.array([[3.43896131, -2.08938436]])))
    self.assertIs(activation, cActivation)

  def test__propagate_forward(self):
    X = np.array([
      [-0.31178367, 0.72900392, 0.21782079, -0.8990918 ],
      [-2.48678065, 0.91325152, 1.12706373, -1.51409323],
      [1.63929108, -0.4298936, 2.63128056, 0.60182225],
      [-0.33588161, 1.23773784, 0.11112817, 0.12915125],
      [0.07612761, -0.15512816, 0.63422534, 0.810655]
    ])

    parameters = {
      'W1': np.array([
        [0.35480861, 1.81259031, -1.3564758 , -0.46363197, 0.82465384],
        [-1.17643148, 1.56448966, 0.71270509, -0.1810066 , 0.53419953],
        [-0.58661296, -1.48185327, 0.85724762, 0.94309899, 0.11444143],
        [-0.02195668, -2.12714455, -0.83440747, -0.46550831, 0.23371059]
      ]),
      'b1': np.array([
        [1.38503523],
        [-0.51962709],
        [-0.78015214],
        [0.95560959]
      ]),
      'W2': np.array([
        [-0.12673638, -1.36861282, 1.21848065, -0.85750144],
        [-0.56147088, -1.0335199 , 0.35877096, 1.07368134],
        [-0.37550472, 0.39636757, -0.47144628, 2.33660781]
      ]),
      'b2': np.array([
        [1.50278553],
        [-0.59545972],
        [0.52834106]
      ]),
      'W3': np.array([
        [0.9398248, 0.42628539, -0.75815703]
      ]),
      'b3': np.array([[-0.16236698]])
    }

    AL, caches = _propagate_forward(parameters, X)
    self.assertTrue(np.allclose(AL, np.array([[0.03921668, 0.70498921, 0.19734387, 0.04728177]])))
    self.assertEqual(len(caches), 3)
  
  def test__step_backward(self):
    dAL = np.array([[-0.41675785, -0.05626683]])
    A_prev = np.array([
      [-2.1361961, 1.64027081],
      [-1.79343559, -0.84174737],
      [0.50288142, -1.24528809]
    ])
    W = np.array([[-1.05795222, -0.90900761, 0.55145404]])
    b =  np.array([[2.29220801]])
    Z = np.array([[0.04153939, -1.11792545]])
    
    cache = ((A_prev, W, b), (Z, 'sigmoid'))
    dA_prev, dW, db = _step_backward(dAL, cache)

    self.assertTrue(np.allclose(dA_prev, np.array([[0.11017994, 0.01105339], [0.09466817, 0.00949723], [-0.05743092, -0.00576154]])))
    self.assertTrue(np.allclose(dW, np.array([[0.10266786, 0.09778551, -0.01968084]])))
    self.assertTrue(np.allclose(db, [[-0.05729622]]))
    
    cache = ((A_prev, W, b), (Z, 'relu'))
    dA_prev, dW, db = _step_backward(dAL, cache)

    self.assertTrue(np.allclose(dA_prev, np.array([[0.44090989, 0.], [0.37883606, 0.], [-0.2298228, 0.]])))
    self.assertTrue(np.allclose(dW, np.array([[0.44513824, 0.37371418, -0.10478989]])))
    self.assertTrue(np.allclose(db, [[-0.20837892]]))

  def test__propagate_back(self):
    AL = np.array([[1.78862847, 0.43650985]])
    Y = np.array([[1, 0]])
    A_prev = np.array([
      [0.09649747, -1.8634927],
      [-0.2773882 , -0.35475898],
      [-0.08274148, -0.62700068],
      [-0.04381817, -0.47721803]
    ])
    W = np.array([
      [-1.31386475,  0.88462238,  0.88131804,  1.70957306],
      [0.05003364, -0.40467741, -0.54535995, -1.54647732],
      [0.98236743, -1.10106763, -1.18504653, -0.2056499]
    ])
    b = np.array([
      [1.48614836],
      [0.23671627],
      [-1.02378514]
    ])
    Z = np.array([
      [-0.7129932 ,  0.62524497],
      [-0.16051336, -0.76883635],
      [-0.23003072,  0.74505627]
    ])
    first_cache = ((A_prev, W, b), (Z, 'relu'))

    A_prev = np.array([
      [1.97611078, -1.24412333],
      [-0.62641691, -0.80376609],
      [-2.41908317, -0.92379202]
    ])
    W = np.array([[-1.02387576,  1.12397796, -0.13191423]])
    b = np.array([[-1.62328545]])
    Z = np.array([[0.64667545, -0.35627076]])
    
    second_cache = ((A_prev, W, b), (Z, 'sigmoid'))
    
    caches = [first_cache, second_cache]

    expected = {
      'dW1': np.array([
        [ 0.41010002, 0.07807203, 0.13798444, 0.10502167],
        [0.,0.,0.,0.],
        [0.05283652, 0.01005865, 0.01777766, 0.0135308]
      ]),
      'db1': np.array([[-0.22007063], [0.], [-0.02835349]]),
      'dW2': np.array([[-0.39202432, -0.13325855, -0.04601089]]),
      'db2': np.array([[ 0.15187861]])
    }

    grads = _propagate_back(AL, Y, caches)

    self.assertTrue(np.allclose(expected['dW1'], grads['dW1']))
    self.assertTrue(np.allclose(expected['db1'], grads['db1']))
    self.assertTrue(np.allclose(expected['dW2'], grads['dW2']))
    self.assertTrue(np.allclose(expected['db2'], grads['db2']))

  def test__cost(self):
    Y, A = np.array([[1, 1, 1]]), np.array([[0.8, 0.9, 0.4]])
    self.assertEqual(_cost(A, Y), 0.414931599615397)
  
  def test_propagate(self):
    w, b, X, Y = np.array([[1., 2.]]), 2., np.array([[1.,2.,-1.],[3.,4.,-3.2]]), np.array([[1,0,1]])
    parameters = {
      'W1': w,
      'b1': b
    }
    grads, cost = propagate(parameters, X, Y)
    self.assertTrue(np.allclose(grads['dW1'], np.array([[0.99845601, 2.39507239]])))
    self.assertTrue(np.allclose(grads['db1'], np.array([[0.00145558]])))
    self.assertEqual(cost, 5.801545319394553)



if __name__ == '__main__':
  unittest.main()


import numpy as np

def gradient_descent(parameters, gradients, alpha=0.1):
  result = {}
  L = len(parameters) // 2
  for l in range(L):
    W = "W{}".format(str(l+1))
    b = "b{}".format(str(l+1))
    result[W] = parameters[W] - alpha * gradients['d{}'.format(W)]
    result[b] = parameters[b] - alpha * gradients['d{}'.format(b)]

  return result


import numpy as np

def initial_velocity(parameters):
  """
  Initializes the velocity as a python dictionary with:
              - keys: "dW1", "db1", ..., "dWL", "dbL" 
              - values: numpy arrays of zeros of the same shape as the corresponding gradients/parameters.
  Arguments:
  parameters -- python dictionary containing your parameters.
                  parameters['W' + str(l)] = Wl
                  parameters['b' + str(l)] = bl
  
  Returns:
  v -- python dictionary containing the current velocity.
                  v['dW' + str(l)] = velocity of dWl
                  v['db' + str(l)] = velocity of dbl
  """
  
  L = len(parameters) // 2 # number of layers in the neural networks
  v = {}
  
  # Initialize velocity
  for l in range(L):
      v["dW" + str(l+1)] = np.zeros(parameters["W{}".format(l+1)].shape)
      v["db" + str(l+1)] = np.zeros(parameters["b{}".format(l+1)].shape)
      
  return v

def update_parameters(parameters, gradients, v, beta, learning_rate):
  """
  Update parameters using Momentum
  
  Arguments:
  parameters -- python dictionary containing your parameters:
                  parameters['W' + str(l)] = Wl
                  parameters['b' + str(l)] = bl
  gradients -- python dictionary containing your gradients for each parameters:
                  gradients['dW' + str(l)] = dWl
                  gradients['db' + str(l)] = dbl
  v -- python dictionary containing the current velocity:
                  v['dW' + str(l)] = ...
                  v['db' + str(l)] = ...
  beta -- the momentum hyperparameter, scalar
  learning_rate -- the learning rate, scalar
  
  Returns:
  parameters -- python dictionary containing your updated parameters 
  v -- python dictionary containing your updated velocities
  """

  L = len(parameters) // 2 # number of layers in the neural networks

  for l in range(L):    
    # compute velocities
    v["dW" + str(l+1)] = beta*v['dW{}'.format(l+1)] + (1-beta) * gradients['dW{}'.format(l+1)]
    v["db" + str(l+1)] = beta*v['db{}'.format(l+1)] + (1-beta) * gradients['db{}'.format(l+1)]
    # update parameters
    parameters["W" + str(l+1)] = parameters["W" + str(l+1)] - learning_rate * v['dW{}'.format(l+1)]
    parameters["b" + str(l+1)] = parameters["b" + str(l+1)] - learning_rate * v['db{}'.format(l+1)]
        
  return parameters, v



import numpy as np
try:
  from activations import sigmoid, sigmoid_derivative, relu, relu_derivative
except ImportError:
  from .activations import sigmoid, sigmoid_derivative, relu, relu_derivative

def propagate(parameters, X, Y, activation="sigmoid"):
  assert(parameters['W1'].shape[1] == X.shape[0])
  assert(X.shape[1] == Y.shape[1])

  A, caches = _propagate_forward(parameters, X, activation)
  
  Y = Y.reshape(A.shape) # ensure Y is the same shape as AL
  
  grads = _propagate_back(A, Y, caches)
  return grads, _cost(A, Y)

def _propagate_forward(parameters, A_prev, activation="relu", output_activation="sigmoid"):
  caches = []
  L = len(parameters) // 2
  
  # Input + hidden layers
  for l in range(1, L):
    W = parameters['W{}'.format(l)]
    b = parameters['b{}'.format(l)]

    A_prev, cache = _step_forward(A_prev, W, b, activation)
    caches.append(cache)

  # Output layer
  W = parameters['W{}'.format(L)]
  b = parameters['b{}'.format(L)]
  AL, cache = _step_forward(A_prev, W, b, output_activation)
  caches.append(cache)
  
  return AL, caches

def _step_forward(A_prev, W, b, activation="sigmoid"):
  # linear
  # Z[l] = W[l] dot A[l-1] + b[l]
  Z = np.dot(W, A_prev) + b
  linear_cache = (A_prev, W, b)
  
  # activation
  # A[l] = g[l](Z[l])
  if activation == "sigmoid":
    A = sigmoid(Z)
  elif activation == "relu":
    A = relu(Z)
  elif activation == "tanh":
    A = np.tanh(Z)
  
  activation_cache = (Z, activation)

  return A, (linear_cache, activation_cache)

def _cost(A, Y):
  m = Y.shape[1]
  logprobs = np.multiply(Y, np.log(A)) + np.multiply((1 - Y), np.log(1 - A))
  cost = -np.sum(logprobs) / m
  cost = np.squeeze(cost) # Turns [[17]] to 17
  return cost

def _propagate_back(AL, Y, caches):
  grads = {}
  L = len(caches) # number of layers  

  # derivative of the cost function with respect to AL
  dA = -(np.divide(Y, AL) - np.divide(1 - Y, 1 - AL))

  # Loop from l=L-1 to l=0
  for l in reversed(range(L)):
    current_cache = caches[l]
    dw, db = "dW{}".format(l + 1), "db{}".format(l + 1)
    dA, grads[dw], grads[db] = _step_backward(dA, current_cache)
  
  return grads


def _step_backward(dA, cache):
  ((A_prev, W, b), (Z, activation)) = cache

  # dZ = dA * g[l]'(Z[l])
  if activation == "relu":
    activation_derivative = relu_derivative(Z)
  elif activation == "sigmoid":
    activation_derivative = sigmoid_derivative(Z)

  dZ = dA * activation_derivative

  m = A_prev.shape[1]
  dA_prev = np.dot(np.transpose(W), dZ) # * D / keep_prob
  dW = np.divide(np.dot(dZ, np.transpose(A_prev)), m)# + (lambd/m) * W
  db = np.divide(np.sum(dZ, axis=1, keepdims=True), m)

  return dA_prev, dW, db


import numpy as np
import math

def random_mini_batches(X, Y, mini_batch_size = 64):
  """
  Creates a list of random minibatches from (X, Y)
  
  Arguments:
  X -- input data, of shape (input size, number of examples)
  Y -- true "label" vector (1 for blue dot / 0 for red dot), of shape (1, number of examples)
  mini_batch_size -- size of the mini-batches, integer
  
  Returns:
  mini_batches -- list of synchronous (mini_batch_X, mini_batch_Y)
  """

  assert(X.ndim == 2)
  assert(Y.ndim == 2)
  assert(X.shape[1] == Y.shape[1])
      
  (shuffled_X, shuffled_Y) = _shuffle(X, Y)
  return _partition(shuffled_X, shuffled_Y, mini_batch_size)

  
def _shuffle(X, Y):
  m = X.shape[1]  # number of training examples

  permutation = list(np.random.permutation(m))
  shuffled_X = X[:, permutation]
  shuffled_Y = Y[:, permutation].reshape((1,m))
  return (shuffled_X, shuffled_Y)

def _partition(shuffled_X, shuffled_Y, mini_batch_size):
  m = shuffled_X.shape[1]
  mini_batches = []
  num_complete_minibatches = math.floor(m/mini_batch_size) # number of mini batches of size mini_batch_size in your partitionning
  
  for k in range(0, num_complete_minibatches):
      mini_batch_X = shuffled_X[:, k*mini_batch_size:(k+1)*mini_batch_size]
      mini_batch_Y = shuffled_Y[:, k*mini_batch_size:(k+1)*mini_batch_size]
      mini_batch = (mini_batch_X, mini_batch_Y)
      mini_batches.append(mini_batch)
  
  # Handling the end case (last mini-batch < mini_batch_size)
  if m % mini_batch_size != 0:
      mini_batch_X = shuffled_X[:, num_complete_minibatches*mini_batch_size:]
      mini_batch_Y = shuffled_Y[:, num_complete_minibatches*mini_batch_size:]
      mini_batch = (mini_batch_X, mini_batch_Y)
      mini_batches.append(mini_batch)

  return mini_batches


import numpy as np

def sigmoid(Z):
  return np.power(1 + np.exp(-Z), -1)

def sigmoid_derivative(Z):
  S = sigmoid(Z)
  return S * (1 - S)

def relu(Z):
  return np.maximum(0,Z)

def relu_derivative(Z):
  return (Z > 0).astype(int)


import numpy as np

def initialize(parameters) :
  """
  Initializes v and s as two python dictionaries with:
              - keys: "dW1", "db1", ..., "dWL", "dbL" 
              - values: numpy arrays of zeros of the same shape as the corresponding gradients/parameters.
  
  Arguments:
  parameters -- python dictionary containing your parameters.
                  parameters["W" + str(l)] = Wl
                  parameters["b" + str(l)] = bl
  
  Returns: 
  v -- python dictionary that will contain the exponentially weighted average of the gradient.
                  v["dW" + str(l)] = ...
                  v["db" + str(l)] = ...
  s -- python dictionary that will contain the exponentially weighted average of the squared gradient.
                  s["dW" + str(l)] = ...
                  s["db" + str(l)] = ...

  """
  
  L = len(parameters) // 2 # number of layers in the neural networks
  v = {}
  s = {}
  
  # Initialize v, s. Input: "parameters". Outputs: "v, s".
  for l in range(L):
      v["dW" + str(l+1)] = np.zeros(parameters["W{}".format(l+1)].shape)
      v["db" + str(l+1)] = np.zeros(parameters["b{}".format(l+1)].shape)
      s["dW" + str(l+1)] = np.zeros(parameters["W{}".format(l+1)].shape)
      s["db" + str(l+1)] = np.zeros(parameters["b{}".format(l+1)].shape)
  
  return v, s

def update_parameters(parameters, grads, v, s, t, learning_rate = 0.01,
                                beta1 = 0.9, beta2 = 0.999,  epsilon = 1e-8):
    """
    Update parameters using Adam
    
    Arguments:
    parameters -- python dictionary containing your parameters:
                    parameters['W' + str(l)] = Wl
                    parameters['b' + str(l)] = bl
    grads -- python dictionary containing your gradients for each parameters:
                    grads['dW' + str(l)] = dWl
                    grads['db' + str(l)] = dbl
    v -- Adam variable, moving average of the first gradient, python dictionary
    s -- Adam variable, moving average of the squared gradient, python dictionary
    learning_rate -- the learning rate, scalar.
    beta1 -- Exponential decay hyperparameter for the first moment estimates 
    beta2 -- Exponential decay hyperparameter for the second moment estimates 
    epsilon -- hyperparameter preventing division by zero in Adam updates

    Returns:
    parameters -- python dictionary containing your updated parameters 
    v -- Adam variable, moving average of the first gradient, python dictionary
    s -- Adam variable, moving average of the squared gradient, python dictionary
    """
    
    L = len(parameters) // 2                 # number of layers in the neural networks
    v_corrected = {}                         # Initializing first moment estimate, python dictionary
    s_corrected = {}                         # Initializing second moment estimate, python dictionary
    
    # Perform Adam update on all parameters
    for l in range(L):
      dWl = "dW{}".format(l+1)
      dbl = "db{}".format(l+1)

      # Moving average of the gradients. Inputs: "v, grads, beta1". Output: "v".
      v[dWl] = beta1 * v[dWl] + (1-beta1) * grads[dWl]
      v[dbl] = beta1 * v[dbl] + (1-beta1) * grads[dbl]

      # Compute bias-corrected first moment estimate. Inputs: "v, beta1, t". Output: "v_corrected".
      v_corrected[dWl] = v[dWl] / (1-np.power(beta1,t))
      v_corrected[dbl] = v[dbl] / (1-np.power(beta1,t))

      # Moving average of the squared gradients. Inputs: "s, grads, beta2". Output: "s".
      s[dWl] = beta2 * s[dWl] + (1-beta2) * np.power(grads[dWl],2)
      s[dbl] = beta2 * s[dbl] + (1-beta2) * np.power(grads[dbl],2)

      # Compute bias-corrected second raw moment estimate. Inputs: "s, beta2, t". Output: "s_corrected".
      s_corrected[dWl] = s[dWl] / (1-np.power(beta2,t))
      s_corrected[dbl] = s[dbl] / (1-np.power(beta2,t))

      # Update parameters. Inputs: "parameters, learning_rate, v_corrected, s_corrected, epsilon". Output: "parameters".
      parameters["W" + str(l+1)] = parameters["W" + str(l+1)] - learning_rate * v_corrected[dWl] / (np.sqrt(s_corrected[dWl]) + epsilon)
      parameters["b" + str(l+1)] = parameters["b" + str(l+1)] - learning_rate * v_corrected[dbl] / (np.sqrt(s_corrected[dbl]) + epsilon)

    return parameters, v, s


import numpy as np

def vectorize(array, m=1):
  """
  Argument:
  image -- a numpy array of shape (a,b, ..., z)
  
  Returns:
  v -- a vector of shape (a*b* ... *z, 1)
  """

  assert(type(array) is np.ndarray)
  return array.reshape(m, -1).T

def normalize_rows(X):
  """    
  Argument:
  X -- A numpy matrix of shape (n, m)
  
  Returns:
  The numpy matrix X normalized (by row) numpy matrix
  """
  # Each row of X is divided by the sqare root of the squared sums of its elements
  # ||[a,b,c]|| == np.sqrt(a**2 + b**2 + c**2)
  X_norm = np.linalg.norm(X, axis=1, ord=2, keepdims=True)

  # Replace zeros to avoid division error
  X_norm[X_norm == 0] = 1
  return X / X_norm

def softmax(X):
  """
  Calculates the softmax for each row of the input x.

  Your code should work for a row vector and also for matrices of shape (n, m).

  Argument:
  x -- A numpy matrix of shape (n,m)

  Returns:
  s -- A numpy matrix equal to the softmax of x, of shape (n,m)
  """
  
  # Array of elements e raised to power of corresponding element in X
  X_exp = np.exp(X)

  # A vector of the sum of each row of x_exp
  X_sum = np.sum(X_exp, axis=1, keepdims=True)
  
  return X_exp / X_sum

def loss(yhat, y, L=2):
  assert(L in [1,2])
  if L == 1:
    return np.sum(np.absolute(y - yhat))
  elif L == 2:
    return np.sum(np.power(y - yhat, 2))


import numpy as np


def initialize_parameters(layer_dimensions, W_multiplier=0.01, use="random", sqrt_multiplier=2, sqrt_nodes_count_exponent=-1):
    """
    Inputs:
        layer_dimensions -- list of dimensions of each layer in network including the input layer

    Output:
        parameters -- dictionary containing W1, b1, ... WL, bL
    """

    parameters = {}

    # number of layers in network
    L = len(layer_dimensions)

    if (L < 2):
        raise ValueError(
            "layer_dimensions must have at least 2 elements (one input, one output")

    for l in range(1, L):
        W = 'W{}'.format(str(l))
        b = 'b{}'.format(str(l))
        current_layer_nodes_count = layer_dimensions[l]
        previous_layer_nodes_count = layer_dimensions[l - 1]

        if use == "zeros":
            parameters[W] = np.zeros((current_layer_nodes_count, previous_layer_nodes_count))
        elif use == "random":
            parameters[W] = np.random.randn(current_layer_nodes_count, previous_layer_nodes_count) * W_multiplier
        elif use == "he":
            parameters[W] = np.random.randn(
                current_layer_nodes_count, previous_layer_nodes_count) * np.sqrt(sqrt_multiplier * (previous_layer_nodes_count ** sqrt_nodes_count_exponent))
        else:
            raise ValueError("invalid 'use' argument")

        parameters[b] = np.zeros((current_layer_nodes_count, 1))

    return parameters


def initialize_lr_with_zeros(num_features):
    if num_features < 1:
        raise ValueError('num_features must be greater than zero')

    return initialize_parameters([num_features, 1], use="zeros")
